--------------------------------------------------------------------------------
--
-- File: newexi.vhd
-- Version: 0.1 Start date 01/07/2010 (Ian Callaghan)
-- Version: 0.2 Start date 27/07/2013 (Steven Taffs)
--
-- Description: 
--
-- USB to EXI
-- Note: Had to remove "4" / "5" / "E" commands to get the code
-- to fit within the CPLD target
--
-- Targeted device: CPLD  xc9572xl-10vq44
-- Author: Steven Taffs (code port of work by Ian Callaghan to cpld target)
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.std_logic_unsigned.all;

-- Uncomment the following library declaration if using
-- arithmetic functions with Signed or Unsigned values
--use IEEE.NUMERIC_STD.ALL;

-- Uncomment the following library declaration if instantiating
-- any Xilinx primitives in this code.
--library UNISIM;
--use UNISIM.VComponents.all;

----------------------------------------------------------------
-- Main entity
----------------------------------------------------------------
entity usbexi is
    Port ( -- USB
           usb_pwren : in  STD_LOGIC;
           usb_txe   : in  STD_LOGIC;
           usb_rxf   : in  STD_LOGIC;
           usb_rd    : out  STD_LOGIC;
           usb_wr    : out  STD_LOGIC;
           usb_data  : inout  STD_LOGIC_VECTOR (7 downto 0);
           -- MISC
           led : inout   STD_LOGIC;
           -- GAMECUBE
           exi_cs  : in   STD_LOGIC;
           exi_clk : in  STD_LOGIC;
           exi_do  : out  STD_LOGIC;
           exi_di  : in   STD_LOGIC
           );
end usbexi;

architecture arch_usbexi of usbexi is

----------------------------------------------------------------
-- signal defines
----------------------------------------------------------------
signal reset_conf : std_logic;                         -- holds inverted usb_pwren state
signal exi_count : std_logic_vector(3 downto 0);       -- counts number of exi bits received
signal exi_cmd : std_logic_vector(3 downto 0);         -- holds command sent from gc
signal id_mode_set : std_logic;                        -- set to one if id command was received
signal led_state : std_logic := '0';                   -- controls LED on / off state
signal usb_tx_status_set : std_logic;                  -- set if we can write data to usb
signal usb_rx_status_set : std_logic;                  -- set if there is data to be read from usb
signal usb_write_mode_set : std_logic;                 -- set when gc wants to write data to usb
signal exi_read_buffer : std_logic_vector(7 downto 0); -- holds data to be written to usb
signal usb_read_mode_set : std_logic;                  -- set when gc wants to read data from usb
signal exi_usb_data_in : std_logic_vector(7 downto 0); -- holds data read from usb

----------------------------------------------------------------
begin
----------------------------------------------------------------
reset_conf <= not usb_pwren;                  -- pwren is high during suspend, low once configured
                                              -- so we flip it to have reset_conf = 0 for suspend

led <= usb_pwren when led_state = '0' else led_state;

----------------------------------------------------------------
-- Main State Machine
----------------------------------------------------------------
process (exi_cs, exi_clk, reset_conf)

begin
   if(exi_cs = '1' or reset_conf = '0') then
        exi_count <= (others => '0');
        exi_cmd <= (others => '0');
        exi_read_buffer <= (others => '0');
        exi_usb_data_in <= (others => '0');
        exi_do <= '0';
        usb_rd <= '1';
        usb_wr <= '0';
		  usb_data(7 downto 0) <= (others => 'Z'); -- best to put here for suspend reasons
        usb_tx_status_set <= '0';
        usb_rx_status_set <= '0';
        usb_write_mode_set <= '0';
        usb_read_mode_set <= '0';
        id_mode_set <= '0';
   else
      if (exi_clk'event and exi_clk ='1') then
      
         case exi_count is
               when "0000" => 
                   exi_cmd(3) <= exi_di;          -- clock in command bit 3 from gc
                   exi_do <= '0';

               when "0001" =>
                   exi_cmd(2) <= exi_di;          -- clock in command bit 2 from gc 
                   exi_do <= '0';

               when "0010" =>
                   exi_cmd(1) <= exi_di;          -- clock in command bit 1 from gc 
                   exi_do <= '0';

               when "0011" =>
                   exi_cmd(0) <= exi_di;          -- clock in command bit 0 from gc 

                   if(exi_cmd = x"A") then        -- 'A' receive byte from PC
                       if(usb_rxf = '0') then           -- is there anything to read?
                           exi_do <= '1';               -- ok tell gc
                           usb_read_mode_set <= '1';
                           usb_rd <= '0';               -- send RD low to get byte  
                       end if;
                   end if;

               when "0100" =>
                   exi_read_buffer(7) <= exi_di;  -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       exi_do <= '0';             -- clear read mode ack
                   end if;

                   if(exi_cmd = x"7") then        -- was '7' LED off command sent ?
                       led_state <= '1';          -- yes, set LED state to off
                   end if;
                
                   if(exi_cmd = x"8") then        -- was '8' LED on command sent ?
                       led_state <= '0';          -- yes, set LED state to on
                   end if;

                   if(exi_cmd = x"9") then        -- was '9' ID command sent ?
                       id_mode_set <= '1';        -- yes, set ID flag
                       exi_do <= '1';             -- send ID bit 0
                   end if;

                   if(exi_cmd = x"B") then        -- was 'B' send byte to PC command sent ?
                       if(usb_txe = '0') then                 -- can we write?
                           exi_do <= '1';                     -- ok tell gc
                           usb_write_mode_set <= '1';
                       end if;
                   end if;

                   if(exi_cmd = x"C") then        -- was 'C' TX status command sent ?
                       if(usb_txe = '0') then                 -- can we write?
                           exi_do <= '1';                     -- ok tell gc
                           usb_tx_status_set <= '1';
                       end if;
                   end if;

                   if(exi_cmd = x"D") then        -- was 'D' RX status command sent ?
                       if(usb_rxf = '0') then                 -- are we read to read?
                           exi_do <= '1';                     -- ok tell gc
                           usb_rx_status_set <= '1';
                       end if;
                   end if;

               when "0101" =>
                   exi_read_buffer(6) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then  
                       exi_usb_data_in(7 downto 0) <= usb_data(7 downto 0);  -- actually get byte from usb here
                   end if;

                   if(id_mode_set = '1') then
                       exi_do <= '0';                 -- send ID bit 1
                   end if;

                   if(usb_write_mode_set = '1' or usb_tx_status_set = '1' or usb_rx_status_set = '1') then
                       exi_do <= '0';                 -- clear TX / RX status or write ack
                       usb_tx_status_set <= '0';
                       usb_rx_status_set <= '0';
                   end if;

               when "0110" =>
                   exi_read_buffer(5) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       usb_rd <= '1';                 -- set read strobe pin back to default high state                         
                   end if;

                   if(id_mode_set = '1') then
                       exi_do <= '0';                 -- send ID bit 2
                   end if;

               when "0111" =>
                   exi_read_buffer(4) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(7);  -- send read usb fifo data to gc
                   end if;

                   if(id_mode_set = '1') then
                       exi_do <= '0';                 -- send ID bit 3
                   end if;

               when "1000" =>
                   exi_read_buffer(3) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(6);  -- send read usb fifo data to gc
                   end if;

                   if(id_mode_set = '1') then
                       exi_do <= '1';                 -- send ID bit 4
                   end if;               

               when "1001" =>
                   exi_read_buffer(2) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(5);  -- send read usb fifo data to gc
                   end if;

                   if(id_mode_set = '1') then
                       exi_do <= '1';                 -- send ID bit 5
                   end if;

               when "1010" => 
                   exi_read_buffer(1) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(4);  -- send read usb fifo data to gc
                   end if;

                   if(id_mode_set = '1') then
                       exi_do <= '1';                 -- send ID bit 6
                   end if;

               when "1011" => 
                   exi_read_buffer(0) <= exi_di;      -- read in data to send to usb fifo

                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(3);  -- send read usb fifo data to gc
                   end if;

                   if(id_mode_set = '1') then 
                       exi_do <= '0';                 -- send ID bit 7
                       id_mode_set <= '1';            -- eight ID bits sent so clear ID flag
                   end if;

               when "1100" =>
                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(2);  -- send read usb fifo data to gc              
                   end if;

                   if(usb_write_mode_set = '1') then
                       usb_wr <= '1';                 -- WR most go high for 20ns minimum before data
                       usb_data(7 downto 0) <= exi_read_buffer;
                   end if;

               when "1101" =>
                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(1);  -- send read usb fifo data to gc
                   end if;

               when "1110" => 
                   if(usb_read_mode_set = '1') then
                       exi_do <= exi_usb_data_in(0);  -- send read usb fifo data to gc
                   end if;

                   if(usb_write_mode_set = '1') then   
                       usb_wr <= '0';                 -- WR goes low to write / send data to usb fifo                
                       usb_data(7 downto 0) <= (others => 'Z');
                       usb_write_mode_set <= '0';
                   end if;

               when others =>                -- default case added to stop the compiler moaning !

         end case;

      exi_count <= exi_count + '1';        -- increment bit counter 

      end if;
   end if;
end process;

end arch_usbexi;

